/*
 * Simple lock manager without queues but busy loops
 *
 * The lock directory is expected to have two subdirectories:
 * - clients       has a file for each client that is named after
 *                 its id and has its id as content
 * - locks         locks are held by hard-linking the client files
 *                 out of the clients directory to a file that
 *                 is named after the lock name
 */

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <afblib/inbuf.h>
#include <afblib/outbuf.h>
#include "lockmanager.h"

/* read the id out of a lock file */
static int read_name(inbuf* ibuf, stralloc* name) {
   int ch;
   if (!stralloc_copys(name, "")) return 0;
   while ((ch = inbuf_getchar(ibuf)) >= 0 && ch != '\n') {
      if (!stralloc_readyplus(name, 1)) return 0;
      name->s[name->len++] = ch;
   }
   return name->len > 0;
}

static int create_path(stralloc* path,
      lockset* set, char* subdir, char* entry) {
   if (!stralloc_copys(path, set->dirname)) return 0;
   if (!stralloc_cats(path, "/")) return 0;
   if (!stralloc_cats(path, subdir)) return 0;
   if (!stralloc_cats(path, "/")) return 0;
   if (!stralloc_cats(path, entry)) return 0;
   if (!stralloc_0(path)) return 0;
   return 1;
}

/*
 * initialize lock set
 */
int lm_init(lockset* set, char* dirname, char* myname) {
   if (!strhash_alloc(&set->locks, 4)) return 0;
   set->dirname = strdup(dirname);
   if (set->dirname == 0) return 0;
   set->myname = strdup(myname);
   if (set->myname == 0) return 0;
   set->myfile = (stralloc) {0};

   if (!create_path(&set->myfile, set, "clients", myname)) return 0;
   int fd = open(set->myfile.s, O_WRONLY|O_CREAT|O_EXCL, 0444);
   if (fd < 0) return 0;
   outbuf obuf = {fd};
   int ok =
      outbuf_write(&obuf,
         set->myname, strlen(set->myname)) == strlen(set->myname) &&
      outbuf_putchar(&obuf, '\n') == '\n' &&
      outbuf_flush(&obuf);
   outbuf_free(&obuf); close(fd);
   if (!ok) {
      unlink(set->myfile.s); return 0;
   }
   return 1;
}

/* release all locks associated with set and allocated storage */
void lm_free(lockset* set) {
   if (strhash_start(&set->locks)) {
      char* lockname;
      while (strhash_next(&set->locks, &lockname)) {
         stralloc path = {0};
         if (create_path(&path, set, "locks", lockname)) {
            unlink(path.s);
            stralloc_free(&path);
         }
      }
   }
   unlink(set->myfile.s);
   stralloc_free(&set->myfile);
   free(set->dirname); set->dirname = 0;
   free(set->myname); set->myname = 0;
   strhash_free(&set->locks);
}

/*
 * check status of the given lock and return
 * the name of the holder in holder if its held
 * and an empty string if the lock is free
 */
int lm_stat(lockset* set, char* lockname, stralloc* holder) {
   if (!stralloc_copys(holder, "")) return 0;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return 0;
   int fd = open(path.s, O_RDONLY);
   stralloc_free(&path);
   if (fd < 0) return 1;
   inbuf ibuf = {fd};
   int ok = read_name(&ibuf, holder);
   close(fd); inbuf_free(&ibuf);
   if (!ok) return 0;
   return 1;
}

/* attempt to lock `lockname' */
int lm_lock(lockset* set, char* lockname) {
   if (!strhash_add(&set->locks, lockname, 0)) return 0;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return 0;
   int ok;
   for(;;) {
      errno = 0;
      if (link(set->myfile.s, path.s) == 0) {
         ok = 1; break;
      }
      if (errno != EEXIST) {
         ok = 0; break;
      }
      sleep(1);
   }
   stralloc_free(&path);
   if (!ok) strhash_remove(&set->locks, lockname);
   return ok;
}

/* attempt to lock `lockname' but do not block */
int lm_nonblocking_lock(lockset* set, char* lockname) {
   if (!strhash_add(&set->locks, lockname, 0)) return 0;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return 0;
   int ok = link(set->myfile.s, path.s) == 0;
   if (!ok) strhash_remove(&set->locks, lockname);
   stralloc_free(&path);
   return ok;
}

/* release `lockname' */
int lm_release(lockset* set, char* lockname) {
   if (!strhash_remove(&set->locks, lockname)) return 0;
   stralloc path = {0};
   if (!create_path(&path, set, "locks", lockname)) return 0;
   int ok = unlink(path.s) == 0;
   stralloc_free(&path);
   return ok;
}
